/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <cray@nyobsd.org> wrote this file. As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.
 *
 * Jean-Philippe Jacoupy
 * ----------------------------------------------------------------------------
 */

#include "postgres.h"
#include "fmgr.h"

#ifdef PG_MODULE_MAGIC
PG_MODULE_MAGIC;
#endif


Datum bytea_or(PG_FUNCTION_ARGS) ;
Datum bytea_and(PG_FUNCTION_ARGS) ;
Datum bytea_xor(PG_FUNCTION_ARGS) ;

PG_FUNCTION_INFO_V1(bytea_or);
Datum
bytea_or( PG_FUNCTION_ARGS ) {

    bytea *_arg1, *_arg2, *_result, *biggest, *lowest;
    int a1size, a2size, maxsize, minsize;
    int i, j, maxbitsize, minbitsize;

    /* No nullity check we will be called as STRICT */
    _arg1 = PG_GETARG_BYTEA_P(0);
    _arg2 = PG_GETARG_BYTEA_P(1);

    a1size = VARSIZE(_arg1);
    a2size = VARSIZE(_arg2);

    if (a1size <= a2size) {
        biggest = _arg2;
        lowest = _arg1;
        maxsize = a2size;
        minsize = a1size;
    }
    else {
        biggest = _arg1;
        lowest = _arg2;
        maxsize = a1size;
        minsize = a2size;
    }

    maxbitsize = (maxsize - VARHDRSZ) * 8;
    minbitsize = (minsize - VARHDRSZ) * 8;

    _result = (bytea *)palloc(maxsize);
    memset(_result, 0, maxsize);
    SET_VARSIZE(_result, maxsize);

    j = 0;
    for(i = 0; i < maxbitsize; i++) {
        if( (i != 0) && (i % 8 == 0) ) {
            j++;
        }
        if (i < minbitsize) {
            _result->vl_dat[j] |= ( ( ( ( ( biggest->vl_dat[j] ) >> (i % 8) ) & 1 ) | ( ( ( lowest->vl_dat[j] ) >> (i % 8) ) & 1 ) ) << (i % 8) );
        }
        else {
            _result->vl_dat[j] |= ( ( ( ( ( biggest->vl_dat[j] ) >> (i % 8) ) & 1 ) | 0 ) << (i % 8) );
        }
    }

    PG_RETURN_BYTEA_P(_result);
}

PG_FUNCTION_INFO_V1(bytea_and);
Datum
bytea_and( PG_FUNCTION_ARGS ) {

    bytea *_arg1, *_arg2, *_result, *biggest, *lowest;
    int a1size, a2size, maxsize, minsize;
    int i, j, maxbitsize, minbitsize;

    /* No nullity check we will be called as STRICT */
    _arg1 = PG_GETARG_BYTEA_P(0);
    _arg2 = PG_GETARG_BYTEA_P(1);

    a1size = VARSIZE(_arg1);
    a2size = VARSIZE(_arg2);

    if (a1size <= a2size) {
        biggest = _arg2;
        lowest = _arg1;
        maxsize = a2size;
        minsize = a1size;
    }
    else {
        biggest = _arg1;
        lowest = _arg2;
        maxsize = a1size;
        minsize = a2size;
    }

    maxbitsize = (maxsize - VARHDRSZ) * 8;
    minbitsize = (minsize - VARHDRSZ) * 8;

    _result = (bytea *)palloc(maxsize);
    memset(_result, 0, maxsize);
    SET_VARSIZE(_result, maxsize);

    j = 0;
    for(i = 0; i < maxbitsize; i++) {
        if( (i != 0) && (i % 8 == 0) ) {
            j++;
        }
        if (i < minbitsize) {
            _result->vl_dat[j] |= ( ( ( ( ( biggest->vl_dat[j] ) >> (i % 8) ) & 1 ) & ( ( ( lowest->vl_dat[j] ) >> (i % 8) ) & 1 ) ) << (i % 8) );
        }
        else {
            _result->vl_dat[j] |= ( ( ( ( ( biggest->vl_dat[j] ) >> (i % 8) ) & 1 ) & 1 ) << (i % 8) );
        }
    }

    PG_RETURN_BYTEA_P(_result);
}

PG_FUNCTION_INFO_V1(bytea_xor);
Datum
bytea_xor( PG_FUNCTION_ARGS ) {

    bytea *_arg1, *_arg2, *_result, *biggest, *lowest;
    int a1size, a2size, maxsize, minsize;
    int i, j, maxbitsize, minbitsize;

    /* No nullity check we will be called as STRICT */
    _arg1 = PG_GETARG_BYTEA_P(0);
    _arg2 = PG_GETARG_BYTEA_P(1);

    a1size = VARSIZE(_arg1);
    a2size = VARSIZE(_arg2);

    if (a1size <= a2size) {
        biggest = _arg2;
        lowest = _arg1;
        maxsize = a2size;
        minsize = a1size;
    }
    else {
        biggest = _arg1;
        lowest = _arg2;
        maxsize = a1size;
        minsize = a2size;
    }

    maxbitsize = (maxsize - VARHDRSZ) * 8;
    minbitsize = (minsize - VARHDRSZ) * 8;

    _result = (bytea *)palloc(maxsize);
    memset(_result, 0, maxsize);
    SET_VARSIZE(_result, maxsize);

    j = 0;
    for(i = 0; i < maxbitsize; i++) {
        if( (i != 0) && (i % 8 == 0) ) {
            j++;
        }
        if (i < minbitsize) {
            _result->vl_dat[j] |= ( ( ( ( ( biggest->vl_dat[j] ) >> (i % 8) ) & 1 ) ^ ( ( ( lowest->vl_dat[j] ) >> (i % 8) ) & 1 ) ) << (i % 8) );
        }
        else {
            _result->vl_dat[j] |= ( ( ( ( ( biggest->vl_dat[j] ) >> (i % 8) ) & 1 ) ^ 0 ) << (i % 8) );
        }
    }

    PG_RETURN_BYTEA_P(_result);
}
